/* *** Parser-mania. (c) 1994 Antonio J. Antn. Tlf 96-5881219 *** */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

enum codis { numero, texte, string, operador, separador, final,
	     error, obrir, tancar, paraula_reservada, identificador,
	     node_unari, node_binari, node_dades, UNKNOWN };

enum tipus { fnc, op, cte };

#define FALSE		(0)
#define TRUE 		(!FALSE)

#define MAX_VARS 	100
#define MAX_TOKEN	32
#define MAX_STRING	255
#define MAX_DATA	128

#define New(x)		((p##x) malloc(sizeof(x)))
#define IsFunction(x)	(IsType((x), fnc))
#define IsConstant(x)	(IsType((x), cte))
#define IsOperator(x)	(IsType((x), op))
#define IsError()	(errortext[0] != 0)
extern unsigned _stklen = 16384;	/* NOU TAMANY DE LA PILA */

typedef int bool;

typedef struct
{
  int tipus;
  int llargada;
  union
   {
     char operador[3];
     char *string;
     double numero;
     char ident[MAX_TOKEN];
     int codi_paraula;
   } dada;
} item;

typedef struct
{
  int  tipus_node;
  item element;
  void *esquerra;
  void *dreta;
} node;

typedef struct
{
  char ident[MAX_TOKEN];
  item valor;
} variable;

typedef struct
{
  char *ident;
  int  que_es;
  int  parametres;
} reservat;

typedef item     *pitem;
typedef node     *pnode;
typedef variable *pvariable;
typedef reservat *preservat;

pvariable variables[MAX_VARS];
int       ultima_var = 0;
double    parser_result = 0.0;
item      ultim_element;
bool      ultim_valid = FALSE;

reservat reserved[] =
{
  { "rnd",   fnc, 0 }, { "acos",  fnc, 1 },  { "asin",  fnc, 1 }, /*  1.. 3 */
  { "atan",  fnc, 1 }, { "atan2", fnc, 2 },  { "ceil",  fnc, 1 }, /*  4.. 6 */
  { "cos",   fnc, 1 }, { "cosh",  fnc, 1 },  { "exp",   fnc, 1 }, /*  7.. 9 */
  { "floor", fnc, 1 }, { "mod",   op,  0 },  { "ln",    fnc, 1 }, /* 10..12 */
  { "log",   fnc, 1 }, { "pow",   fnc, 2 },  { "pow10", fnc, 1 }, /* 13..15 */
  { "sin",   fnc, 1 }, { "sinh",  fnc, 1 },  { "sqrt",  fnc, 1 }, /* 16..18 */
  { "tan",   fnc, 1 }, { "tanh",  fnc, 1 },  { "pi",    cte, 0 }, /* 19..21 */
  { "ans",   cte, 0 }, { "and",   op,  0 },  { "or",    op,  0 }, /* 22..24 */
  { "not",   op,  0 }, { "xor",   op,  0 },  { "u",     fnc, 1 }, /* 25..27 */
  { "delta", fnc, 1 }, { "min",   fnc, -1},  { "max",   fnc, -1}, /* 28..30 */
  { "vlist", fnc, 0 }, { "sinc",  fnc, 1 },  { "randomize", fnc, 1}, /* 31..33 */
  { "sum",   fnc, 4 }, { "prod",  fnc, 4 },  { "sgn",   fnc, 1 }, /* 34..36 */
  { NULL,    0,   0 },
};

/* **** FUNCIONS D'ERROR ***** */

char errortext[80];

/* **** FUNCIO D'ERROR **** */

int errorf(char *message)
{
  if(errortext[0] == 0)
   {
     strcpy(errortext, message);
     return TRUE;
   }
  return FALSE;
}

/* **** NETEJAR ELS ERRORS **** */

int clearerror(void)
{
  errortext[0] = 0;
  return TRUE;
}

/* **** OBTINDRE EL MISSATGE D'ERROR **** */

char *GetErrorText(void)
{
  return errortext;
}

/* ****** FUNCIONS D'ENTRADA DE DADES ****** */

/* **** FUNCIONS D'ENTRADA BASICA DE CARACTERS **** */

int  mygetc_pos;
char mygetc_data[MAX_DATA];

/* *** INICIALITZAR AMB EL STRING *** */

bool initmygetc(char *texte)
{
  mygetc_pos = 0;
  strcpy(mygetc_data, texte);

  return TRUE;
}

/* *** OBTINDRE CARACTER *** */

int mygetc(FILE *device)
{
  return mygetc_data[mygetc_pos++]; 		/*  getc(device) */
}

/* *** RETORNAR CARACTER *** */

int unmygetc(int c, FILE *device)
{
  return (mygetc_data[--mygetc_pos] = c); 	/* ungetc(c, device) */
}

/* ****** FUNCIO QUE RETORNA UN ITEM A LA ENTRADA ***** */

int ungetitem(pitem element)
{
  memcpy(&ultim_element, element, sizeof(item));
  ultim_valid = TRUE;

  return element->tipus;
}

/* ****** FUNCIO QUE OBTE UN ITEM DE LA ENTRADA ***** */
/* ****** VERSIO QUE NO PERMET STRINGS **** */

int getitem(pitem element)
{
  char codi, codi2, tmp[MAX_TOKEN];
  int pos = 0, tipus = -1;
  bool fin = FALSE;

  if(ultim_valid)
   {
     ultim_valid = FALSE;
     memcpy(element, &ultim_element, sizeof(item));
     return ultim_element.tipus;
   }

  codi = mygetc(stdin);
  while(!fin)
   {
     switch(codi)
      {
	case ' ' : if(pos > 0)
		      fin = TRUE;
	case '\r': break;
	case 0xFF:
	case 0x00:
	case '\n': if(pos > 0)
		    unmygetc(codi, stdin);
		   else
		    tipus = final;
		   fin = TRUE;
		   break;
	case ',' : if(pos > 0)
		    unmygetc(codi, stdin);
		   else
		    tipus = separador;
		   fin = TRUE;
		   break;
	case '.' : if(tipus == numero)
		    {
		      tmp[pos] = 0x0;
		      if(strchr(tmp, '.') != NULL)
		       {
			 errorf("Numero mal construit");
			 fin = TRUE;
			 tipus = error;
		       }
		      tmp[pos++] = codi;
		    }
		   else
		    {
		      tipus = numero;
		      tmp[pos++] = codi;
		      codi2 = mygetc(stdin);
		      if(!isdigit(codi2))
		       {
			 errorf("Numero mal construit");
			 fin = TRUE;
			 tipus = error;
		       }
		      unmygetc(codi2, stdin);
		    }
		   break;
	case '(':
	case ')':
	case '[':
	case ']':
	case '+':
	case '-':
	case '*':
	case '/':
	case '^':
	case '~':
	case '&':
	case '|':
	case '!':
	case '?':
	case ':':
	case '=':
	case '>':
	case '<':
	case '%':  if(pos > 0) {
		      unmygetc(codi, stdin);
		      fin = TRUE; }
		   else {
		      tmp[pos++] = codi;
		      switch(codi)
		       { case '(' : tipus = obrir; break;
			 case ')' : tipus = tancar; break;
			 default  : tipus = operador;
				    codi2 = mygetc(stdin);
				    switch(codi)
				     {
				       case ':' :
				       case '>' :
				       case '<' :
				       case '!' :
				       case '=' : if(codi2 == '=')
						   tmp[pos++] = codi2;
						  else
						   unmygetc(codi2,stdin);
						  break;
				       default  : unmygetc(codi2, stdin);
				     }
		       }
		      fin = TRUE; }
		   break;
	default:   if(pos == 0) {
		    if(isdigit(codi))
		     tipus = numero;
		    else
		     tipus = texte;
		    tmp[pos++] = codi;
		    break; }

		   switch(tipus)
		    {
		      case numero : if(!isdigit(codi)) {
				       errorf("Numero mal creat");
				       fin = TRUE;
				       tipus = error;
				       break; }
		      case texte  : tmp[pos++] = codi;
				    break;
		      default	  : errorf("Error del analitzador lexic");
				    tipus = error;
				    fin = TRUE;
		    }
      }
     if(pos == MAX_TOKEN - 1) {
	errorf("Element massa llarg");
	tipus = error;
	fin = TRUE; }
     if(!fin)
      codi = mygetc(stdin);
  }
  tmp[pos] = 0x0;

  switch(tipus)
   {
     case numero   : sscanf(tmp, "%lf", &element->dada.numero); break;
     case operador : strcpy(element->dada.operador, tmp); break;
     case texte    : element->dada.codi_paraula = IsReservedWord(tmp);
		     if(element->dada.codi_paraula)
		      tipus = paraula_reservada;
		     else
		      strcpy(element->dada.ident, tmp);
   }
  element->tipus    = tipus;
  element->llargada = pos;
  return tipus;
}

/* ***** FUNCIO PER DETERMINAR SI ES PARAULA RESERVADA ***** */

int IsReservedWord(char *ident)
{
  int n = 0;

  while(reserved[n].ident != NULL)
   {
     if(stricmp(ident, reserved[n++].ident) == 0)
      return n;
   }

  return 0;
}

/* **** VERIFICACIO DEL TIPUS DE PARAULA RESERVADA **** */

bool IsType(int index, int tipus)
{
  return (reserved[index - 1].que_es == tipus);
}

/* ****** FUNCIONS DE CREACIO DE NODES ******* */

/* ****** CREA UN NODE BINARI ***** */

pnode NodeBinari(pitem element, pnode esquerra, pnode dreta)
{
  pnode n;

  if(esquerra == NULL || dreta == NULL)
   return NULL;

  n = New(node);
  n->tipus_node = node_binari;
  memcpy(&(n->element), element, sizeof(item));
  n->esquerra = esquerra;
  n->dreta = dreta;

  return n;
}

/* ****** CREA UN NODE UNARI ****** */

pnode NodeUnari(pitem element, pnode operand)
{
  pnode n;

  if(operand == NULL)
   return NULL;

  n = New(node);
  n->tipus_node = node_unari;
  memcpy(&(n->element), element, sizeof(item));
  n->esquerra = operand;
  n->dreta = NULL;

  return n;
}

/* ****** NODE AMB DADES (TERMINAL) ***** */

pnode NodeDada(pitem element, int tipus)
{
  pnode n = New(node);

  n->tipus_node = tipus;
  memcpy(&(n->element), element, sizeof(item));
  n->dreta = NULL;
  n->esquerra = NULL;

  return n;
}

/* ******* PARSER ******* */

pnode expr(void)
{
  pnode n, expr1();
  item element;

  getitem(&element);
  if(element.tipus == final)
   return NULL;
  ungetitem(&element);

  n = expr1();
  getitem(&element);
  if(element.tipus != final)
   errorf("Final de expresi esperada");

   return n;
}

/* ****** expr1 ::= expr11 | expr11 > expr1 | expr11 < expr1 ***** */
pnode expr1(void)
{
  pnode n, m, expr11();
  item  element;

  n = expr11();

  if(n != NULL)
   for(;;)
    {
      getitem(&element);
      if(element.tipus != operador || (element.dada.operador[0] != '>' &&
	 element.dada.operador[0] != '<'))
       break;
      m = expr11();
      n = NodeBinari(&element, n, m);
    }
  ungetitem(&element);
  return n;
}

/* ****** expr11 ::= expr2 | expr2 + expr11 | expr2 - expr11 ***** */

pnode expr11(void)
{
  pnode n, m, expr2();
  item  element;

  n = expr2();

  if(n != NULL)
   for(;;)
    {
      getitem(&element);
      if(element.tipus != operador || (element.dada.operador[0] != '+' &&
	 element.dada.operador[0] != '-'))
       break;
      m = expr2();
      n = NodeBinari(&element, n, m);
    }
  ungetitem(&element);
  return n;
}

/* ***** expr2 ::= expr3 | expr3 * expr2 | expr3 / expr2 |
		   expr3 mod expr2 ***** */

pnode expr2(void)
{
  pnode n, m, expr3();
  item  element;

  n = expr3();

  if(n != NULL)
   for(;;)
    {
      getitem(&element);
      if((element.tipus != operador || (element.dada.operador[0] != '*' &&
	  element.dada.operador[0] != '/')) && (element.tipus != paraula_reservada ||
	  element.dada.codi_paraula != 11))
       break;
      m = expr3();
      n = NodeBinari(&element, n, m);
    }
  ungetitem(&element);
  return n;
}

/* **** expr3 ::= expr4 | expr4 ^ expr3 ***** */

pnode expr3(void)
{
  pnode n, expr4();
  item  element;

  n = expr4();

  if(n != NULL)
   {
     getitem(&element);
     if(element.tipus == operador && element.dada.operador[0] == '^')
      return NodeBinari(&element, n, expr3());
     else
      ungetitem(&element);
   }
  return n;
}

/* **** expr4 ::= constant | numero | -expr4 | (expr1) | simbol = expr1 |
		  ident(llista)  **** */

pnode expr4(void)
{
  item  element;
  pnode n, llista();

  getitem(&element);
  if(element.tipus == numero)
   return NodeDada(&element, numero);
  else if(element.tipus == texte)
   {
     item element2;

     getitem(&element2);
     if(element2.tipus == operador && element2.dada.operador[0] == '=' &&
	element2.llargada == 1)
      return NodeBinari(&element2, NodeDada(&element, texte), expr1());
     else
      ungetitem(&element2);
     return NodeDada(&element, texte);
   }
  else if(element.tipus == paraula_reservada && IsConstant(element.dada.codi_paraula))
   return NodeDada(&element, paraula_reservada);
  else if(element.tipus == paraula_reservada && IsFunction(element.dada.codi_paraula))
   {
     item element2;

     getitem(&element2);
     if(element2.tipus != obrir)
      {
	errorf("Parentesis per obrir esperat");
	return NULL;
      }

     getitem(&element2);		/* Funcio sense parametres */
     if(element2.tipus == tancar)
      return NodeDada(&element, paraula_reservada);
     ungetitem(&element2);

     n = NodeUnari(&element, llista());
     getitem(&element2);
     if(element2.tipus != tancar)
      {
	errorf("Parentesis per tancar esperat");
	return NULL;
      }
   }
  else if(element.tipus == operador && element.dada.operador[0] == '-')
   return NodeUnari(&element, expr4());
  else if(element.tipus == obrir)
   {
     n = expr1();
     getitem(&element);
     if(element.tipus != tancar)
      errorf("Esperat parentesis per tancar");
   }
  else
   {
     errorf("Expresio incorrecta");
     n = NULL;
   }

  return n;
}

/* **** llista ::= expr1 | expr1, llista **** */

pnode llista(void)
{
  pnode n;
  item element;

  n = expr1();
  getitem(&element);
  if(element.tipus == separador)
   n = NodeBinari(&element, n, llista());
  else
   ungetitem(&element);

  return n;
}

/* ***** FUNCIO QUE DONA EL NODE DEL PARAMETRE N-ESIM **** */
/* (per funcions de ms d'un parmetre) */

pnode parametre(pnode n, int ordre)
{
  if(ordre == 1)
   return (pnode) n->esquerra;
  else if(ordre == 2)
   if(((pnode)n->dreta)->element.tipus == separador)
    return parametre((pnode) n->dreta, 1);
   else
    return (pnode) n->dreta;
  else
   return parametre((pnode) n->dreta, ordre - 1);
}

/* **** FUNCIO QUE RETORNA EL NRO DE PARAMETRES **** */

unsigned nroparametres(pnode n)
{
  pnode    k;
  unsigned p = 0;

  k = (pnode) n->esquerra;
  if(k != NULL)
   p +=1;
  while(k->dreta != NULL)
   {
     k = (pnode) k->dreta;
     p += 1;
   }

  return p;
}

/* ****** EVALUADOR ****** */

double eval(pnode n)
{
  double op1, op2, GetVariable(), SetVariable();

  switch(n->tipus_node)
   {
     case node_binari : if(n->element.tipus == paraula_reservada)
		    switch(n->element.dada.codi_paraula)
		     {
		       case 11 : op1 = eval((pnode) n->esquerra);
				 op2 = eval((pnode) n->dreta);
				 return fmod(op1, op2);
		       default : errorf("Construcci incorrecta");
				 return 0.0;
		     }
		    else
		   switch(n->element.dada.operador[0])
		    {
		      case '>' :
		      case '<' :
		      case '^' :
		      case '+' :
		      case '-' :
		      case '*' :
		      case '/' : op1 = eval((pnode) n->esquerra);
				 op2 = eval((pnode) n->dreta);
				 switch(n->element.dada.operador[0])
				  {
				    case '+' : return op1 + op2;
				    case '-' : return op1 - op2;
				    case '*' : return op1 * op2;
				    case '/' : return op1 / op2;
				    case '^' : return pow(op1,op2);
				    case '>' : return op1 > op2;
				    case '<' : return op1 < op2;
				  }
		      case '=' : return SetVariable(((pnode)n->esquerra)->element.dada.ident,
						    eval(n->dreta));
		    }
		   break;
     case node_unari :
		   if(n->element.tipus == paraula_reservada)
		    {
		      switch(n->element.dada.codi_paraula)
		       {
		       case 29 :
		       case 30 :
				 {
				   double val;
				   int l = nroparametres(n), k;
				   if(l < 2)
				    {
				      errorf("Numero de parametres insuficient");
				      return 0.0;
				    }
				   op1 = eval(parametre(n->esquerra, 1));
				   op2 = op1;
				   for(k = 2; k <= l; k++)
				    {
				      val = eval(parametre(n->esquerra, k));
				      op1 = min(op1, val);
				      op2 = max(op2, val);
				    }
				   if(n->element.dada.codi_paraula == 29)
				    return op1;
				   else
				    return op2;
				 }
			 case 34 :
			 case 35 : {
				     double acum, inici, final, cont;
				     pnode  var, fnc;
				     acum = (n->element.dada.codi_paraula == 34) ? 0 : 1;

				     if(nroparametres(n) != 4)
				      {
					errorf("Numero de parametres incorrecte");
					return 0.0;
				      }
				     inici = eval(parametre(n->esquerra, 3));
				     final = eval(parametre(n->esquerra, 4));
				     fnc   = parametre(n->esquerra, 1);
				     var   = parametre(n->esquerra, 2);

				     for(cont = inici; cont <= final; cont++)
				      {
					SetVariable(var->element.dada.ident, cont);
					if(n->element.dada.codi_paraula == 34)
					 acum += eval(fnc);
					else
					 acum *= eval(fnc);
				      }

				     return acum;
				   }
			 case  5 :
			 case 14 :  op1 = eval(parametre(n->esquerra, 1));
				    op2 = eval(parametre(n->esquerra, 2));
				    return n->element.dada.codi_paraula == 5 ?
					   atan2(op1, op2) : pow(op1, op2);
			 case 2 : return acos(eval(n->esquerra));
			 case 3 : return asin(eval(n->esquerra));
			 case 4 : return atan(eval(n->esquerra));
			 case 6 : return ceil(eval(n->esquerra));
			 case 7 : return cos(eval(n->esquerra));
			 case 8 : return cosh(eval(n->esquerra));
			 case 9 : return exp(eval(n->esquerra));
			 case 10: return floor(eval(n->esquerra));
			 case 11: errorf("Construcci incorrecta");
				  return 0.0;
			 case 12: return log(eval(n->esquerra));
			 case 13: return log10(eval(n->esquerra));
			 case 15: return pow10(eval(n->esquerra));
			 case 16: return sin(eval(n->esquerra));
			 case 17: return sinh(eval(n->esquerra));
			 case 18: return sqrt(eval(n->esquerra));
			 case 19: return tan(eval(n->esquerra));
			 case 20: return tanh(eval(n->esquerra));
			 case 27: return (eval(n->esquerra) >= 0) ? 1 : 0;
			 case 28: return (eval(n->esquerra) == 0) ? 1 : 0;
			 case 32: op1 = eval(n->esquerra);
				  return (op1 == 0) ? 1 : sin(op1)/op1;
			 case 33: srand(op1 = eval(n->esquerra));
				  return op1;
			 case 36: return 2.0*(eval(n->esquerra) >= 0) - 1.0;
			 default: errorf("Funcio no implementada");
				  return 0.0;
		       }
		    }
		   else if(n->element.dada.operador[0] == '-')
		    return -eval((pnode) n->esquerra);
		   break;
     case paraula_reservada:
		   switch(n->element.dada.codi_paraula)
		    {
		      case  1 : return (double) rand() / RAND_MAX;
		      case 21 : return atan(1)*4;
		      case 22 : return parser_result;
		      case 31 : if(VariablesDefined())
				 {
				   int k;

				   printf("Llistat de variables definides:\n");
				   for(k = 0; k < ultima_var; k++)
				    printf("%32s = %lf\n", variables[k]->ident, GetVariable(variables[k]->ident));
				 }
				return ultima_var;
		      default : if(IsFunction(n->element.dada.codi_paraula))
				 errorf("Funcio requereix de parametres");
				else
				 errorf("Constant no definida");
				return 0.0;
		    }
     case texte  : return GetVariable(n->element.dada.ident);
     case numero : return n->element.dada.numero;
   }
  errorf("Error del evaluador");
  return 0.0;
}

/* ******** LLIBERA LA MEMORIA NODAL UTILITZADA ****** */

void lliberar_memoria(pnode n)
{
  if(n->tipus_node == node_unari)
   lliberar_memoria(n->esquerra);
  else if(n->tipus_node == node_binari) {
   lliberar_memoria(n->esquerra);
   lliberar_memoria(n->dreta); }
  free(n);
}

/* ***** FUNCIONS D'ACCES A VARIABLES ****** */

bool VariablesDefined(void)
{
  return (ultima_var != 0);
}

double GetVariable(char *ident)
{
  int k;

  for(k = 0; k < ultima_var; k++)
   if(stricmp(variables[k]->ident, ident) == 0)
     return variables[k]->valor.dada.numero;
  errorf("Variable no definida");
  return 0.0;
}

double SetVariable(char *ident, double valor)
{
  int k;
  variable temp;

  for(k = 0; k < ultima_var; k++)
   {
     if(stricmp(variables[k]->ident, ident) == 0)
      {
	variables[k]->valor.dada.numero = valor;
	return valor;
      }
   }
  strcpy(temp.ident, ident);
  temp.valor.tipus = numero;
  temp.valor.dada.numero = valor;
  variables[ultima_var] = (pvariable) malloc(sizeof(variable));
  memcpy(variables[ultima_var], &temp, sizeof(variable));
  ultima_var++;
  return valor;
}

bool ReleaseVariable(char *ident)
{
  int k, l;

  for(k = 0; k < ultima_var; k++)
   if(stricmp(variables[k]->ident, ident) == 0)
    {
      free(variables[k]);
      for(l = k + 1; l < ultima_var; l++)
       memcpy(variables[l - 1], variables[l], sizeof(variable));
      ultima_var -= 1;
      return TRUE;
    }
  errorf("Variable no definida");
  return FALSE;
}

/* ****** SECUENCIADOR PARSER-EVALUADOR ****** */

double Evaluar(char *tExpr)
{
  pnode n;

  clearerror();
  initmygetc(tExpr);
  n = expr();
  if(n != NULL)
   {
     parser_result = eval(n);
     lliberar_memoria(n);
   }

  return (IsError()) ? 0 : parser_result;
}

